
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include "qcmatch.h"

#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif

typedef struct {
	REAL	coord;
} NODE;

REAL	euclid_cost( void *nx, void *ny );

BOOLEAN
Circle_Omega ( MNODE *x1, MNODE *x2, MNODE *x3, void **input, COSTFUNC cost,
	int psi );

static void Circle_Get_Xover( REAL x1, REAL y1, REAL x2, REAL y2,
		REAL delta, REAL *x3, REAL *y3 );

COSTFUNC	CostFunc  = euclid_cost;
OMEGAFUNC	OmegaFunc = Circle_Omega;

REAL
euclid_cost( void *nx, void *ny )
{
	REAL   x1, x2, y1, y2;

	x1 = cos(((NODE *)nx)->coord);
	y1 = sin(((NODE *)nx)->coord);

	x2 = cos(((NODE *)ny)->coord);
	y2 = sin(((NODE *)ny)->coord);

	return sqrt((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2));
}

BOOLEAN
Circle_Omega ( MNODE *x1, MNODE *x2, MNODE *x3, void **input, COSTFUNC cost,
	int psi )
{
	REAL a1,b1;	/* x-y coordinates of MNODE x1 */
	REAL a2,b2;	/* x-y coordinates of MNODE x2 */
	REAL a3,b3;	/* x-y coordinates of MNODE x3 */
	REAL a4,b4;	/* x-y coordinates of crossover of x1 and x2 */
	REAL a5,b5;	/* x-y coordinates of crossover of x1 and x2 */

	REAL c4,d4;	/* x-y coordinates for rotation of (a4,b4) */
	REAL c5,d5;	/* x-y coordinates for rotation of (a5,b5) */

	REAL del, eps;	/* Delta and Epsilon values for crossovers */


	a1 = cos(((NODE **)input)[x1->x]->coord);
	b1 = sin(((NODE **)input)[x1->x]->coord);
	a2 = cos(((NODE **)input)[x2->x]->coord);
	b2 = sin(((NODE **)input)[x2->x]->coord);
	a3 = cos(((NODE **)input)[x3->x]->coord);
	b3 = sin(((NODE **)input)[x3->x]->coord);

	del = psi * (x1->i - x2->i);
	eps = psi * (x2->i - x3->i);

	Circle_Get_Xover(a1, b1, a2, b2, del, &a4, &b4);
	Circle_Get_Xover(a2, b2, a3, b3, eps, &a5, &b5);


	/* Rotate the crossover points w.r.t. MNODE x3, and then determine
	 * the relative ordering of x3 and the two crossover points.  For
	 * this, interpret crossover (a4,b4)=(c4,d4) as occurring in the
	 * interval [x2,x1] and the second crossover (a5,b5)=(c5,d5) as
	 * occurring in the interval [x3,x2].
	 */ 
	c4 = a4*a2 + b4*b2;
	d4 = b4*a2 - a4*b2;
	c5 = a5*a2 + b5*b2;
	d5 = b5*a2 - a5*b2;

	if (d4 >= 0) {
		return (d5 <= 0) ? TRUE : (c5 <= c4);

	} else {
		return (d5 > 0) ? FALSE : (c4 <= c5);
	}
}

#define norm(a,b) (sqrt((a)*(a) + (b)*(b)))

void
Circle_Get_Xover( REAL x1, REAL y1, REAL x2, REAL y2, REAL delta,
		  REAL *x3, REAL *y3 )
{
	REAL a; 		/* Half of delta */
	REAL c; 		/* Half of distance from (x1,y1) to (x2,y2) */
	REAL d; 		/* Distance from origin to line through
				   (x1,y1) and (x2,y2) */
	REAL u,v; 		/* (u,v) coordinates of the crossover point */
	REAL alpha,beta;  	/* Midpt between (x1,y1) and (x2,y2) on circle */


	REAL hip; 		/* Half of inner product */
	REAL asqr, csqr;  	/* Intermediate values, for clarity */


	a = 0.5 * delta;

	hip = (x1*x2 + y1*y2)/2;
	if (hip > 0.5) {
		hip = 0.5;
	} else if (hip < -0.5) {
		hip = -0.5;
	}

	c = sqrt(0.5 - hip);
	d = sqrt(0.5 + hip);

	if (-x2*y1 + x1*y2 < 0) {
		d = -d;
	}

	asqr = a*a;
	csqr = c*c;

	if (asqr > csqr) {
		if ((-a) >= 0) {	/* if (-a >= c) */
			*x3 = x1;
			*y3 = y1;

		} else {	   	/* if ( a >= c) */
			*x3 = x2;
			*y3 = y2;
		}
		return;
	}

	u = -(1+d) * (1-(asqr/csqr));
	v = (asqr == csqr) ? a : (a * sqrt(1+(u*u)/(csqr-asqr)));

	if (hip > 0) {
		alpha = (x1 + x2)/(2*d);
		beta  = (y1 + y2)/(2*d);
	} else {
		alpha = (y2 - y1)/(2*c);
		beta  = (x1 - x2)/(2*c);
	}

	*x3 = (u+d)*alpha -     v*beta;
	*y3 =     v*alpha + (u+d)*beta;

	/*
	assert(fabs(norm(*x3,*y3) - 1.0) < 1E-6);
	assert(fabs(-norm(*x3-x1,*y3-y1) + norm(*x3-x2,*y3-y2) + delta) < 1E-6);
	*/
}
