#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <limits.h>
#include <assert.h>
#include "qcmatch.h"

static int	Max_node;	/* Memory is allocated for tours of up to this
				   size */

static BOOLEAN	 LineLike;	/* Whether the tour is line-like */
static OMEGAFUNC Omega;		/* The Omega predicate */
static COSTFUNC	 Cost;		/* The user-supplied cost function */

#define COST(nodex, nodey)	((nodex == NULL || nodey == NULL) ?	\
				 (REAL) 0.0 : (*Cost)(nodex, nodey))

static int	Psi;		/* The color toggle, 1=BLUE, -1=RED */
static int	Xn;		/* The X of Algorithm 2 */
static REAL	TotalCost;	/* The cost of an optimal matching  */

static void **	Nodes;		/* Working copy of the input nodes array */

/* Linked lists for reducing tour to alternating subtours */

static int     *LevelHeads, *LevelTails, *LevelLinks;

/* The M deque is implemented via the following structure.  Its size
   is 3*Max_node */

typedef struct {
	MNODE *r;	/* Points to element M_{R} */
	MNODE *l;	/* Points to element M_{L} */
	MNODE *d ;	/* Vector of MNODE elements */
	int rx;		/* Index of the deque's rightmost element */
	int lx;		/* Index of the deque's leftmost element */
} MDEQUE;

static MDEQUE M;

/* The two left deques L^{1} and L^{-1} are implemented via the
   structure below.  Each contains 3*Max_node entries.	*/

typedef struct {
	MNODE *r;	/* Points to element L_{R} */
	MNODE *rm1;	/* Points to element L_{R-1} */
	MNODE *l;	/* Points to element L_{L} */
	MNODE **d;	/* Vector of pointers into the M deque,
			   comprising each L deque */
	int rx;		/* Index of the deque's rightmost element */
	int lx;		/* Index of the deque's leftmost element */
} LDEQUE;

static LDEQUE *L;

				/* Local procedures */
static void	qc_match_alt( int level, int matching[] );
static void	process_node( int matching[] );
static void	qc_match_alloc( int toursize );
static void    *qc_match_malloc( int bytes );
static void	push_L_right( int c, MNODE *de);
static void	push_M_right( int xx );
static void	pop_L_left( int c );
static int	pop_M_left( void );
static void	pop_L_right( int c );
static int	pop_M_right( void );
static void	match_pair( int matching[] );
static BOOLEAN	generic_omega( MNODE *iLrm, MNODE *iLr, MNODE *iMr,
			      void *input[], COSTFUNC cost, int psi);


REAL
qc_match( int n, void *input[], int colors[], int matching[], BOOLEAN linelike,
			COSTFUNC cost, OMEGAFUNC omega )
{
	int	x, d, level;
	int	min_level, max_level;


	/*  Step 1:  Initialization.  */

	LineLike = linelike;
	Cost	 = cost;
	Omega    = omega ? omega : generic_omega;

	if (n > Max_node) {
		qc_match_free();
		qc_match_alloc(n);
	}

	memcpy(Nodes, input, n*sizeof(void *));
	Nodes[n] = NULL;	/* "Extra" node for any odd-length subtours */


	/*  Step 2:  Read input colors and break into levels. */

	d = 0;			/* d(0) */
	min_level = INT_MAX;
	max_level = INT_MIN;
	memset((char *) LevelLinks,           -1, (  Max_node+1)*sizeof(int));
	memset((char *)(LevelHeads-Max_node), -1, (2*Max_node+1)*sizeof(int));

	for (x=0 ; x < n ; x++) {

	    /* Calculate level(x), and then d(x+1) for next time. */

		if (colors[x] == BLUE) {
			level = d;
			d--;
		} else {	/* RED */
			level = d + 1;
			d++;
		}

	    /* Keep high and low watermarks for level */

		if (level < min_level) {
			min_level = level;
		}
		if (level > max_level) {
			max_level = level;
		}

	    /* Add node x to the tail of the linked list for its level */

		if (LevelHeads[level] == -1){
			LevelHeads[level] = x;
		} else {
			LevelLinks[LevelTails[level]] = x;
		}

		LevelTails[level] = x;
	}

	for (level = min_level; level <= max_level; level++) {

	    /* If subtour is odd-length, add the "extra" node of index n
	       to the level.  LevelLinks[n] will always be -1.
	     */
		if (colors[LevelHeads[level]] == colors[LevelTails[level]]) {
			assert(LineLike);
			LevelLinks[LevelTails[level]] = n;
		}
	}

	/*  Step 3:  Do matching for each level.  */

	TotalCost = 0.0;

	for (level = min_level; level <= max_level; level++) {
		qc_match_alt(level, matching);
	}

	return TotalCost;
}


static void
qc_match_alt( int level, int matching[] )
{
				/* "Initialization" */
	Psi = -1;

	M.lx = 0;
	M.rx = -1;

	L[RED].lx = 0;
	L[RED].rx = -1;

	L[BLUE].lx = 0;
	L[BLUE].rx = -1;

				/* "Read Input into the M Deque" */
	Xn = LevelHeads[level];
	do {
		push_M_right(Xn);
		Psi = -Psi;
	}
	while ((Xn = LevelLinks[Xn]) != -1);

				/* "The First Scan" */
	while ((L[Psi].rx < L[Psi].lx) || (M.l->x != L[Psi].l->x)) {
		Xn = pop_M_left();
		process_node(matching);
		push_L_right(Psi, M.r);
		Psi = -Psi;
	}

				/* "The Second Scan" */
	if (!LineLike) {
		while ((L[RED].rx >= L[RED].lx) || (L[BLUE].rx >= L[BLUE].lx)) {

			Xn = pop_M_left();
			if (Xn == L[Psi].l->x) {
				pop_L_left(Psi);
			}

			process_node(matching);
			Psi = -Psi;
		}
	}

				/* "Windup Processing" */
	while (M.lx <= M.rx) {
		match_pair(matching);
	}
}

static void
process_node( int matching[] )
{
	push_M_right(Xn);

	if (L[-Psi].rx > L[-Psi].lx) {

		while ((L[-Psi].rx > L[-Psi].lx) &&
		       ((COST(Nodes[L[-Psi].rm1->x], Nodes[M.r->x]) - 
			 COST(Nodes[L[-Psi].r  ->x], Nodes[M.r->x])) <
			(Psi*(L[-Psi].r->i - L[-Psi].rm1->i)))) {

				pop_L_right(-Psi);
		}
	}

	if ((L[-Psi].rx >= L[-Psi].lx) &&
	    (COST(Nodes[M.r->x], Nodes[L[-Psi].r->x]) <
	     Psi*(M.r->i - L[-Psi].r->i))) {

		Xn = pop_M_right();

		while (M.r->x != L[-Psi].r->x) {
			match_pair(matching);
		}

		push_M_right(Xn);
	}

	if (LineLike && Nodes[M.r->x] == NULL) {
		return;
	}

						/* >= 2 elements there */
	while ((L[Psi].rx > L[Psi].lx) &&
	       ((*Omega)(L[Psi].rm1, L[Psi].r, M.r, Nodes, Cost, Psi) == TRUE)) {

		pop_L_right(Psi);
	}
}

static void
push_L_right( int c, MNODE *de)
{
	L[c].d[++L[c].rx] = de;

	L[c].rm1 = L[c].r;
	L[c].r	 = de;

	/* Arrange for "l" to be set after first push */

	if (L[c].rx  == L[c].lx) {
		L[c].l = de;
	}
}

static int
pop_M_left( void )
{
	++M.lx;
	return (M.l++)->x;
}

static void
pop_L_left( int c )
{
	++L[c].lx;

	L[c].l = L[c].d[L[c].lx];
}

static int
pop_M_right( void )
{
	--M.rx;
	return (M.r--)->x;
}

static void
pop_L_right( int c )
{
	--L[c].rx;

	L[c].r	 = L[c].rm1;
	L[c].rm1 = L[c].d[L[c].rx-1];
}

/* This is the push_M_right() procedure from the paper */

static void
push_M_right( int x )
{
	REAL   i;

    /* Compute the "Improvement" function */

	if (M.rx >= M.lx) {
		i = M.r->i + Psi * COST(Nodes[M.r->x], Nodes[x]);
	} else {
		i = 0.0;
	}

    /* Push right the node and improvement value onto the MAIN deque */

	M.r = &M.d[++M.rx];
	M.r->x = x;
	M.r->i = i;

	if (M.rx == M.lx) {
		M.l = M.r;
	}
}

static void
match_pair( int matching[] )
{
	int	a, b;

	a = (M.r - 1)->x;
	b = pop_M_right();

	if (Nodes[a]) {
		matching[a] = Nodes[b] ? b : -1;
	}
	if (Nodes[b]) {
		matching[b] = Nodes[a] ? a : -1;
	}

				/* The next line is not in the paper. It
				 * accumulates the total cost.
				 */
	TotalCost += COST(Nodes[a], Nodes[b]);

	if ((L[Psi].rx >= L[Psi].lx) && (M.r->x == L[Psi].r->x)) {
		pop_L_right(Psi);
	}

	pop_M_right();
}

static void
qc_match_alloc( int toursize )
{
	Max_node = toursize;

	Nodes	    = (void **)	qc_match_malloc((Max_node+1) * sizeof(void *));

	LevelLinks  = (int *)	qc_match_malloc((  Max_node+1) * sizeof(int));
	LevelTails  = (int *)	qc_match_malloc((2*Max_node+1) * sizeof(int));
	LevelHeads  = (int *)	qc_match_malloc((2*Max_node+1) * sizeof(int));
	M.d	   = (MNODE *)	qc_match_malloc(3 * (Max_node+1) * sizeof(MNODE));
	L	   = (LDEQUE *) qc_match_malloc(3 * sizeof(LDEQUE));
	L[0].d	   = (MNODE **) qc_match_malloc(3 * (Max_node+1) * sizeof(MNODE *));
	L[2].d	   = (MNODE **) qc_match_malloc(3 * (Max_node+1) * sizeof(MNODE *));

	LevelTails += Max_node;
	LevelHeads += Max_node;
	L += 1;
}

void
qc_match_free( void )
{
	if (Max_node <= 0)
		return;

	L -= 1;
	LevelTails -= Max_node;
	LevelHeads -= Max_node;

	free(L[2].d);
	free(L[0].d);
	free(L);
	free(M.d);
	free(LevelHeads);
	free(LevelTails);
	free(LevelLinks);

	Max_node = 0;
}

static void *
qc_match_malloc( int bytes )
{
	void *p;

	p = malloc(bytes);

	if (p == NULL) {
		fprintf(stderr, "Cannot allocate enough memory!\n");
		exit(-1);
	}

	return(p);
}


static BOOLEAN
generic_omega( MNODE *iLrm, MNODE *iLr, MNODE *iMr, void *input[],
	       COSTFUNC cost, int psi)
{
	int	ilow, ihigh, k;
	BOOLEAN pastxoverA, pastxoverB;
	void	*np_Lrm, *np_Lr, *np_Mr, *np_Mk;

	ilow = M.lx;
	ihigh = (iLrm - &M.d[0]) + 1;  /* Index of the node after iLrm */

				/* Get the node pointers for the three nodes */
	np_Lrm = input[iLrm->x];
	np_Lr  = input[iLr->x];
	np_Mr  = input[iMr->x];

				/* Search M-deque entries M[ilow] through
				 * M[ihigh-2] to find which crossover comes
				 * first.
				 */
	while (ihigh > ilow + 1) {
		k = ilow + 2 * ((ihigh - ilow) >> 2);
		np_Mk = input[M.d[k].x];      /* Precompute M[k] */

				/* The following boolean variable is TRUE
				 * if M[k] is past the L(sub R-1)(sup psi),
				 * L(sub R)(sup psi) crossover point.
				 */
		pastxoverA = (COST(np_Lrm, np_Mk) - COST(np_Lr, np_Mk) <
			      psi * (iLrm->i - iLr->i));

				/* The following boolean variable is TRUE
				 * if M[k] is past the L(sub R)(sup psi),
				 * M(sub R) crossover point.
				 */
		pastxoverB = (COST(np_Lr, np_Mk) - COST(np_Mr, np_Mk) <
			      psi * (iLr->i - M.r->i));

		if (pastxoverA) {
			if (pastxoverB) {
				ihigh = k;
			} else {
				return TRUE;
			}
		} else {
			if (pastxoverB) {
				return FALSE;
			} else {
				ilow = k + 2;
			}
		}
	}
				/* Returning either TRUE or FALSE at this
				 * point will find a minimum-cost matching.
				 */
	return TRUE;
}

