
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <sys/times.h>
#include "qcmatch.h"

extern void	srand48(long);
extern long	lrand48(void);
extern double	drand48(void);

#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif

#define ROUNDOFF	.000005

#define MAX(A,B) ( ((A)>(B)) ? (A) : (B) )
#define MIN(A,B) ( ((A)<(B)) ? (A) : (B) )

				/* Macros for returning a pseudo-random
				 * integer between 0 and max-1.
				 */
#define SRAND(s)	(srand48(s))
#define RAND(max)	(lrand48() % (max))

typedef struct {
	REAL	coord;
} NODE;

int		MaxNodes   = 128;
int		NumTours   = 1;
unsigned	RandomSeed = 102;

REAL	MaxCoord = 2.0*M_PI;

REAL	BYCost;			/* Total cost according to B-Y algorithm */
REAL	DPCost;			/* Total cost according to dynamic programming */

int	TourSize, TourExcess;
int	*Colors, *BYMatching, *Excess;
NODE	**Nodes;

REAL	**DPArray;		/* Dynamic programming array */
int	*DPMatching, *Verify;

int	Successes 		= 0;
int	DifferentMatching 	= 0;
int	Failures 		= 0;
BOOLEAN MoreThanOneSolution;

char	TourType[4]	   = "BRC";
BOOLEAN BalancedTours  	   = TRUE;
BOOLEAN AlternatingTours   = FALSE;
BOOLEAN	LineLike	   = FALSE;

BOOLEAN BenchmarkMode	   = FALSE;
BOOLEAN DynamicProgramming = FALSE;
BOOLEAN GraphicsOutput	   = FALSE;
BOOLEAN ForceGenericOmega  = FALSE;

char	*ProgName;

extern COSTFUNC		CostFunc;
extern OMEGAFUNC	OmegaFunc;

				/* Local procedures */
void	process_options( int argc, char *argv[] );
void	random_test( void );
REAL	dynamic_program( void );
void	dynamic_output( int b, int e );
void	compare_results( void );
REAL	verify_matching( int matching[] );
void	print_matchings( void );
int	coordcmp( const void *n1, const void *n2 );
double	cpu_interval( void );
void *	get_mem( int bytes );

extern void	graphics_output( int, NODE **, int *, int * );

int
main( int argc, char *argv[] )
{
	setbuf(stdout, NULL);

	ProgName = argv[0];

	process_options(argc, argv);

	random_test();

	return 0;
}

void
process_options( int argc, char *argv[] )
{
	extern int	getopt(int argc, char *argv[], char *optstring);
	extern char	*optarg;

	int	opt, err = FALSE;

	while ((opt = getopt(argc, argv, "-BOdgs:t:n:r:")) != -1) {
		switch (opt) {
		case 'B':
			BenchmarkMode = TRUE;
			break;
		case 'O':
			ForceGenericOmega = TRUE;
			break;
		case 'd':
			DynamicProgramming = TRUE;
			break;
		case 'g':
			GraphicsOutput = TRUE;
			break;
		case 's':
			MaxNodes = atoi(optarg);
			if (MaxNodes <= 0) {
				fprintf(stderr, "Bad -n value!\n");
				err = TRUE;
			}
			break;
		case 't':
			strcpy(TourType, optarg);
			if (strlen(TourType) != 3) {
				fprintf(stderr,
					"Tour type must have 3 letters\n");
				err = TRUE;
			}
			break;
		case 'n':
			NumTours = atoi(optarg);
			if (NumTours <= 0) {
				fprintf(stderr, "Bad -n value!\n");
				err = TRUE;
			}
			break;
		case 'r':
			RandomSeed = (unsigned) atoi(optarg);
			break;
		case '?':
			err = TRUE;
			break;
		}
	}

	if (err) {
		fprintf(stderr,	"Usage: %s [-BOdg] [-t {B|U}{R|A}{C|L}] ",
			ProgName);
		fprintf(stderr, "[-s #] [-n #] [-r #]\n");
		fprintf(stderr,
			"\t-B\t\tBenchmark/timing mode (overrides -d,-g)\n");
		fprintf(stderr,
			"\t-O\t\tForce use of generic omega function\n");
		fprintf(stderr,
			"\t-d\t\tVerify with dynamic programming\n");
		fprintf(stderr,
			"\t-g\t\tProduce graphical representation\n");
		fprintf(stderr,
			"\t-t xyz\t\tSpecify 3-letter tour type [BRC]\n");
		fprintf(stderr,
			"\t-s #\t\tSpecify maximum tour size [128]\n");
		fprintf(stderr,
			"\t-n #\t\tSpecify number of problem instances [1]\n");
		fprintf(stderr,
			"\t-r #\t\tSpecify seed for random generator [102]\n");
		exit(1);
	}

				/* Process problem type key */
	switch (TourType[0]) {
	case 'B':	BalancedTours = TRUE;		break;
	case 'U':	BalancedTours = FALSE;		break;
	default:	err = TRUE;			break;
	}

	switch (TourType[1]) {
	case 'R':	AlternatingTours = FALSE;	break;
	case 'A':	AlternatingTours = TRUE;	break;
	default:	err = TRUE;			break;
	}

	switch (TourType[2]) {
	case 'C':	LineLike = FALSE;		break;
	case 'L':	LineLike = TRUE;		break;
	default:	err = TRUE;			break;
	}

	if (!BalancedTours) {
		if (!LineLike) {
			fprintf(stderr, "Unbalanced tours must be line-like!\n");
			err = TRUE;
		}

		if (OmegaFunc && !ForceGenericOmega) {
			fprintf(stderr, "WARNING: Unbalanced tours must ");
			fprintf(stderr, "use generic Omega!\n");
			fprintf(stderr, "Forcing -O ...\n");
			ForceGenericOmega = TRUE;
		}
	}

	if (err) {
		fprintf(stderr, "Illegal tour type\n");
		exit(1);
	}

				/* Cancel -d and -g in benchmark mode */
	if (BenchmarkMode) {
		DynamicProgramming = FALSE;
		GraphicsOutput	   = FALSE;
	}
}

void
random_test( void )
{
	int	node, tour, nblues;
	double	bias, avgtime;


	SRAND(RandomSeed);


    /* ALLOCATE MEMORY NEEDED FOR MAXIMUM TOUR SIZE */

	Colors     = get_mem(MaxNodes * sizeof(int));
	BYMatching = get_mem(MaxNodes * sizeof(int));

	Nodes = get_mem(MaxNodes*sizeof(NODE *));
	for (node=0 ; node < MaxNodes ; node++) {
		Nodes[node] = get_mem(sizeof(NODE));
	}

	if (DynamicProgramming) {
		DPMatching = get_mem( MaxNodes    * sizeof(int));
		Verify     = get_mem( MaxNodes    * sizeof(int));
		Excess     = get_mem((MaxNodes+1) * sizeof(int));

		DPArray = get_mem((MaxNodes+3)*sizeof(REAL *));
		for (node=0 ; node < MaxNodes+3 ; node++) {
			DPArray[node] = get_mem((MaxNodes+3)*sizeof(REAL));
		}
	}


    /* SOLVE MATCHING FOR NumTours PROBLEM INSTANCES */

	avgtime = 0.0;

	for (tour=0 ; tour < NumTours ; tour++) {

	    /* PICK PROBLEM SIZE */

		if (BenchmarkMode) {
				/* Force single tour size */
			TourSize = MaxNodes;

		} else {
			if (BalancedTours) {
				/* Any even number from 0 to 2*(MaxNodes/2) */
				TourSize = 2 * RAND(MaxNodes/2 + 1);

			} else {
				/* Any number from 0 to MaxNodes */
				TourSize = RAND(MaxNodes+1);
			}
		}

	    /* DISTRIBUTE COLORS */

		if (BalancedTours) {

				/* First make all nodes red */
			for (node=0 ; node < TourSize ; node++) {
				Colors[node] = RED;
			}

				/* Alternating: Make nodes 1,3,... blue */
			if (AlternatingTours) {
				for (node=1 ; node < TourSize ; node+=2) {
					Colors[node] = BLUE;
				}

				/* Non-alternating: Make a random half blue */
			} else {
				nblues = 0;
				while (nblues < TourSize / 2) {
					node = RAND(TourSize);

					if (Colors[node] == RED) {
						Colors[node] = BLUE;
						nblues++;
					}
				}
			}

		} else {
				/* Pick colors by flipping a biased coin */
			bias = drand48();

			for (node=0 ; node < TourSize ; node++) {
				Colors[node] = (drand48() <= bias) ? RED : BLUE;
			}
		}

	    /* DISTRIBUTE NODES */

				/* If line-like, restrict circular tours
				 * to the semicircle [0, pi).
				 */
		if (LineLike) {
			MaxCoord = M_PI;
		}
				/* Distribute nodes randomly in the half-open
				 * interval [0,MaxCoord), with resolution 32000.
				 */
		for (node=0 ; node < TourSize ; node++) {
			Nodes[node]->coord = RAND(32000)*(MaxCoord/32000.0);
		}

		qsort(Nodes, TourSize, sizeof(NODE *), coordcmp);

	    /* RUN BUSS-YIANILOS ALGORITHM */

		if (BenchmarkMode) {
			(void) cpu_interval(); 
			BYCost = qc_match(TourSize, (void **)Nodes, Colors,
					  BYMatching, LineLike, CostFunc,
					  ForceGenericOmega ? NULL : OmegaFunc);
			avgtime += (cpu_interval() / NumTours);

		} else {
			BYCost = qc_match(TourSize, (void **)Nodes, Colors,
					  BYMatching, LineLike, CostFunc,
					  ForceGenericOmega ? NULL : OmegaFunc);
		}

	    /* CHECK RESULTS WITH DYNAMIC PROGRAM */

		if (BalancedTours) {
			assert(TourExcess == 0);
		}

		if (DynamicProgramming) {
			DPCost = dynamic_program();
			compare_results();
		}

	    /* DO GRAPHIC OUTPUT */

		if (GraphicsOutput) {
			graphics_output(TourSize, Nodes, Colors, BYMatching);
		}
	}

	if (BenchmarkMode) {
		printf("%6d %10.6f\n", TourSize, avgtime);
	}

	if (DynamicProgramming) {
		fprintf(stderr,
			"\tSuccesses:\t%u (%u with different matchings)\n",
			Successes, DifferentMatching);
		fprintf(stderr, "\tFailures:\t%u\n", Failures);
	}
}

#define SGN(x)			((x) ? (((x) > 0) ? 1 : -1) : 0)
#define SUBPROB(x,y)		((x) <= (y))
#define EXCESS(x,y)		(Excess[(y)+1] - Excess[(x)])
#define EXCESSCOL(x,y)		(SGN(EXCESS((x),(y))))
#define BALANCED(x,y)		(EXCESS((x),(y)) == 0)

#define EDGECOST(x,y)		((*CostFunc)(Nodes[(x)],Nodes[(y)]))
#define MATCHCOST(x,y)		(DPArray[(x)+1][(y)+1])

REAL
dynamic_program( void )
{
	int	i, isize, x, y, xp;
	BOOLEAN	interior, exterior;
	REAL	mincost, cost;

				/* Calculate the balance of the tour */
	Excess[0]  = 0;
	for (i=0 ; i < TourSize ; i++) {
		Excess[i+1] = Excess[i] + Colors[i];
	}
	TourExcess = Excess[TourSize];

				/* "Null intervals" occurring in the algorithm
				 * have the form [x,x-1].  MATCHCOST() should
				 * return zero cost in such cases.
				 */
	for (i=0 ; i <= TourSize ; i++){
		MATCHCOST(i,i-1) = 0.0;
	}


    /* CONSIDER ALL SUBPROBLEM SIZES (1, 2, 3, 4, ..., TourSize) */

	for (isize=1 ; isize <= TourSize ; isize++) {

	    /* CONSIDER SUBPROBLEM INTERVAL AT ALL POSSIBLE POSITIONS [x,y] */

		for (x=0, y=x+isize-1 ; y < TourSize ; x++, y++) {

		    /* CONSIDER ALL MAXIMAL MATCHINGS FOR SUBPROBLEM [x,y]. */

			mincost = HUGE_VAL;

				/* If [x,y] is unbalanced in favor of x's color,
				 * we have to consider the case of unmatched x.
				 */
			if (Colors[x] == EXCESSCOL(x,y)) {
				mincost = MATCHCOST(x+1,y);
			}
			
				/* Loop over all possible matchings for x */
			for (xp=x+1 ; xp <= y ; xp++) {

				if (Colors[x] == Colors[xp]) {
					continue;
				}

				interior = SUBPROB(x+1,xp-1);
				exterior = SUBPROB(xp+1,y);

				if (interior && exterior &&
				    !BALANCED(x+1,xp-1) &&
				    EXCESSCOL(x+1,xp-1) == -EXCESSCOL(xp+1,y)) {
					continue;
				}

				cost = EDGECOST(x, xp);
				if (interior) {
					cost += MATCHCOST(x+1, xp-1);
				}
				if (exterior) {
					cost += MATCHCOST(xp+1, y);
				}

				mincost = MIN(mincost, cost);
			}

			MATCHCOST(x,y) = mincost;
		}
	}

	MoreThanOneSolution = FALSE;

	dynamic_output(0,TourSize-1);

				/* Return cost for complete tour interval */
	return MATCHCOST(0,TourSize-1);
}

void
dynamic_output( int b, int e )		/* Begin, end */
{
	BOOLEAN jumper;
	int	c;


    /* CASE 1: Null problem */
	if (b > e) {
		return;
	}

    /* CASE 2: Single unmatchable point */

	if (b == e) {
		DPMatching[b] = -1;
		return;
	}

    /* CASE 3: Adjacent pair (with possibly identical coordinates) */

	if (e == b+1) {
		if (Colors[b] == Colors[e]) {
			DPMatching[b] = -1;
			DPMatching[e] = -1;
		} else {
			DPMatching[e] = b;
			DPMatching[b] = e;
		}
		return;
	}
		
    /* CASE 4: Arbitrary subproblem of length > 2 */

	jumper = TRUE;

				/* Either b and e are matched with a jumper, or
				 * there is some c such that the cost of 
				 * subproblem [b,c] plus the cost of subproblem
				 * [c+1,e] equals the cost of problem [b,e].
				 * Try to find that c.
				 */
	for (c=b ; c < e ; c++) {

				/* Make sure that the division of the problem
				 * at c does not imply a non-maximal matching.
				 */
		if (!BALANCED(b,c) && EXCESSCOL(b,c) == -EXCESSCOL(c+1,e)) {
			continue;
		}

		if (fabs(MATCHCOST(b,e) - (MATCHCOST(b,c) + MATCHCOST(c+1,e)))
		    <= ROUNDOFF) {

				/* First valid c: Recurse for the subproblems */
			if (jumper) {
				dynamic_output(b, c);
				dynamic_output(c+1, e);
				jumper = FALSE;

				/* Succeeding valid c's: Just set flag to
				 * indicate more than one solution.
				 */
			} else {
				MoreThanOneSolution = TRUE;
			}
		}
	}

	if (jumper) {
				/* If no c was found for this b and e, then
				 * b and e must be matched.  Record this
				 * matching in the matching array...
				 */
		DPMatching[e] = b;
		DPMatching[b] = e;

				/* We know there are intervening nodes, so
				 * recurse on that subproblem.
				 */
		dynamic_output(b+1, e-1);
	}
}

void
compare_results( void )
{
	int	kk;
	BOOLEAN costs_same, outputs_same;
	REAL	byver, dpver;


	byver = verify_matching(BYMatching);
	dpver = verify_matching(DPMatching);

	outputs_same = TRUE;
	for (kk = 0; kk < TourSize; kk++) {
		if (BYMatching[kk] != DPMatching[kk]) {
			outputs_same = FALSE;
		}
	}

	costs_same = (fabs(BYCost - DPCost) < 20*ROUNDOFF);

	if (costs_same) {
		Successes++;
		if (!outputs_same) {
			DifferentMatching++;
		}

	} else {
		Failures++;
		if (outputs_same) {
			printf("COST MISMATCH: Same matchings ???\n");
		} else {
			printf("COST MISMATCH: Different matchings\n");
		}

		print_matchings();
		printf("B-Y Cost: %10f (verify=%f)  DP Cost: %10f (verify=%f)\n",
		       BYCost, byver, DPCost, dpver);

		if (MoreThanOneSolution == TRUE) {
			printf("More than one best solution\n");
		}

		printf("\n");
	}
}

REAL
verify_matching( int matching[] )
{
	int	i, npairs, excess;
	REAL	total_cost;

				/* Copy matching into another buffer */
	for (i=0 ; i < TourSize ; i++) {
		Verify[i] = matching[i];
	}

	total_cost = 0.0;
	npairs     = 0;
	excess     = 0;

	for (i=0 ; i < TourSize ; i++) {

		switch (Verify[i]) {

		case -2:		/* Already-paired node */
			continue;

		case -1:		/* Unmatched node */
			assert(SGN(Colors[i]) == SGN(TourExcess));
			excess += Colors[i];
			continue;

		default:		/* First node of a matched pair */
			assert(Verify[Verify[i]] == i);

			npairs++;
			total_cost += (*CostFunc)(Nodes[i], Nodes[Verify[i]]);

			Verify[Verify[i]] = -2;
			Verify[i] 	  = -2;

			break;
		}
	}

	assert(excess == TourExcess);
	assert(npairs*2 + SGN(excess)*excess == TourSize);

	return total_cost;
}

void
print_matchings( void )
{
	int	jjj, jkj;


	printf("Num nodes = %d\n", TourSize);

	jjj = 0;
	jkj = 0;

	while (1) {
		while (jjj < TourSize && jjj > BYMatching[jjj]){
			jjj++;
		}

		while (jkj < TourSize && jkj > DPMatching[jkj]) {
			jkj++;
		}

		if (jjj >= TourSize || jkj >= TourSize) {
			break;
		}

		printf("%10f  %10f    %10f  %10f\n", Nodes[jjj]->coord,
		       Nodes[BYMatching[jjj]]->coord, Nodes[jkj]->coord,
		       Nodes[DPMatching[jkj]]->coord);

		jjj++;
		jkj++;
	}
}

int
coordcmp( const void *n1, const void *n2 )
{
	REAL	diff;

	diff = (*(NODE **)n1)->coord - (*(NODE **)n2)->coord;

	return (diff > 0.0) ? 1 : ((diff < 0.0) ? -1 : 0);
}

/* Return CPU time in seconds since last call.
   Usage:
		(void) cpu_interval();

		{
			(code to be timed);
		}

		time_elapsed = cpu_interval();
*/

double
cpu_interval( void )
{
        static clock_t	cputime;
        struct tms	tmsbuffer;

        clock_t	thetime, theinterval;


        (void) times(&tmsbuffer);

        thetime     = tmsbuffer.tms_utime;
        theinterval = thetime - cputime;

        cputime = thetime;

        return ((double)theinterval) / sysconf(_SC_CLK_TCK);
}

void *
get_mem( int bytes )
{
	void 	*p;

	if (bytes > 0 && (p = malloc(bytes))) {
		return p;
	} else {
		fprintf(stderr, "%s: malloc failed\n", ProgName);
		exit(1);
	}
}

